import logging
import uuid
from typing import Any, Awaitable, Callable, Dict, Iterable, List, Optional, Text

from convo.core.channels.channel import InputSocket, OutputSocket, UserMsg
import convo.shared.utils.io
from sanic import Blueprint, response
from sanic.request import Request
from sanic.response import HTTPResponse
from socketio import AsyncServer

log = logging.getLogger(__name__)


class ChannelBluePrint(Blueprint):
    def __init__(self, sio: AsyncServer, socketio_path, *args, **kwargs):
        self.sio = sio
        self.socketio_path = socketio_path
        super().__init__(*args, **kwargs)

    def listing(self, app, options) -> None:
        self.sio.attach(app, self.socketio_path)
        super().listing(app, options)


class ChannelIoOutput(OutputSocket):
    @classmethod
    def name(cls) -> Text:
        return "socketio"

    def __init__(self, sio: AsyncServer, bot_message_evt: Text) -> None:
        self.sio = sio
        self.bot_message_evt = bot_message_evt

    async def _sent_msg(self, socket_id: Text, response: Any) -> None:
        """Sends a message to the recipient using the bot event."""

        await self.sio.emit(self.bot_message_evt, response, room=socket_id)

    async def sent_text_msg(
        self, recipient_id: Text, text: Text, **kwargs: Any
    ) -> None:
        """Send a message through this channel."""

        for message_part in text.strip().split("\n\n"):
            await self._sent_msg(recipient_id, {"text": message_part})

    async def sent_img_url(
        self, recipient_id: Text, image: Text, **kwargs: Any
    ) -> None:
        """Sends an image to the output"""

        msg = {"attachment": {"type": "image", "payload": {"src": image}}}
        await self._sent_msg(recipient_id, msg)

    async def sent_text_with_btn(
        self,
        recipient_id: Text,
        text: Text,
        buttons: List[Dict[Text, Any]],
        **kwargs: Any,
    ) -> None:
        """Sends buttons to the output."""

        # split text and create a message for each text fragment
        # the `or` makes sure there is at least one message we can attach the quick
        # replies to
        msg_part = text.strip().split("\n\n") or [text]
        msg = [{"text": message, "quick_replies": []} for message in msg_part]

        # attach all buttons to the last text fragment
        for button in buttons:
            msg[-1]["quick_replies"].append(
                {
                    "content_type": "text",
                    "title": button["title"],
                    "payload": button["payload"],
                }
            )

        for message in msg:
            await self._sent_msg(recipient_id, message)

    async def sent_details(
        self, recipient_id: Text, elements: Iterable[Dict[Text, Any]], **kwargs: Any
    ) -> None:
        """Sends elements to the output."""

        for element in elements:
            message = {
                "attachment": {
                    "type": "template",
                    "payload": {"template_type": "generic", "elements": element},
                }
            }

            await self._sent_msg(recipient_id, message)

    async def sent_modified_dict(
        self, recipient_id: Text, json_message: Dict[Text, Any], **kwargs: Any
    ) -> None:
        """Sends custom json to the output"""

        json_message.setdefault("room", recipient_id)

        await self.sio.emit(self.bot_message_evt, **json_message)

    async def sent_attach(
        self, recipient_id: Text, attachment: Dict[Text, Any], **kwargs: Any
    ) -> None:
        """Sends an attachment to the user."""
        await self._sent_msg(recipient_id, {"attachment": attachment})


class ChannelIoInput(InputSocket):
    """A socket.io input channel."""

    @classmethod
    def name(cls) -> Text:
        return "socketio"

    @classmethod
    def from_cred(cls, credentials: Optional[Dict[Text, Any]]) -> InputSocket:
        credentials = credentials or {}
        return cls(
            credentials.get("user_message_evt", "UserUttered"),
            credentials.get("bot_message_evt", "bot_uttered"),
            credentials.get("namespace"),
            credentials.get("session_persistence", False),
            credentials.get("socketio_path", "/socket.io"),
        )

    def __init__(
        self,
        user_message_evt: Text = "UserUttered",
        bot_message_evt: Text = "bot_uttered",
        namespace: Optional[Text] = None,
        session_persistence: bool = False,
        socketio_path: Optional[Text] = "/socket.io",
    ):
        self.bot_message_evt = bot_message_evt
        self.session_persistence = session_persistence
        self.user_message_evt = user_message_evt
        self.namespace = namespace
        self.socketio_path = socketio_path
        self.sio = None

    def get_output_socket(self) -> Optional["OutputSocket"]:
        if self.sio is None:
            convo.shared.utils.io.raising_warning(
                "SocketIO output channel cannot be recreated. "
                "This is expected behavior when using multiple Sanic "
                "workers or multiple Convo Open Source instances. "
                "Please use a different channel for external events in these "
                "scenarios."
            )
            return
        return ChannelIoOutput(self.sio, self.bot_message_evt)

    def blue_print(
        self, on_new_message: Callable[[UserMsg], Awaitable[Any]]
    ) -> Blueprint:
        # Workaround so that socketio works with requests from others origins.
        # https://github.com/miguelgrinberg/python-socketio/issues/205#issuecomment-493769183
        socket_io = AsyncServer(async_mode="sanic", cors_allowed_origins=[])
        soi_web_hook = ChannelBluePrint(
            socket_io, self.socketio_path, "soi_web_hook", __name__
        )

        # make socket_io object static to use in get_output_channel
        self.sio = socket_io

        @soi_web_hook.route("/", methods=["GET"])
        async def health(_: Request) -> HTTPResponse:
            return response.json({"status": "ok"})

        @socket_io.on("connect", namespace=self.namespace)
        async def connect(sid: Text, _) -> None:
            log.debug(f"User {sid} connected to socketIO endpoint.")

        @socket_io.on("disconnect", namespace=self.namespace)
        async def disconnect(sid: Text) -> None:
            log.debug(f"User {sid} disconnected from socketIO endpoint.")

        @socket_io.on("session_request", namespace=self.namespace)
        async def session_request(sid: Text, data: Optional[Dict]):
            if data is None:
                data = {}
            if "session_id" not in data or data["session_id"] is None:
                data["session_id"] = uuid.uuid4().hex
            if self.session_persistence:
                socket_io.enter_room(sid, data["session_id"])
            await socket_io.emit("session_confirm", data["session_id"], room=sid)
            log.debug(f"User {sid} connected to socketIO endpoint.")

        @socket_io.on(self.user_message_evt, namespace=self.namespace)
        async def handle_msg(sid: Text, data: Dict) -> Any:
            output_channel = ChannelIoOutput(socket_io, self.bot_message_evt)

            if self.session_persistence:
                if not data.get("session_id"):
                    convo.shared.utils.io.raising_warning(
                        "A message without a valid session_id "
                        "was received. This message will be "
                        "ignored. Make sure to set a proper "
                        "session id using the "
                        "`session_request` socketIO event."
                    )
                    return
                sender_id = data["session_id"]
            else:
                sender_id = sid

            message = UserMsg(
                data["message"], output_channel, sender_id, input_channel=self.name()
            )
            await on_new_message(message)

        return soi_web_hook
